unit EasyIpc;

interface

uses
 sysutils, windows, classes, math;

function EasyIpc_Init(const PathName,Options:String):THandle;
function EasyIpc_Free(hIpc:THandle):Boolean;
function EasyIpc_PutText(hIpc:THandle; const Lines:String):Boolean;
function EasyIpc_GetText(hIpc:THandle; Count:Integer):String;

const
 LineEnding      = #13#10;
 DefPipeTimeOut  = 1000;                                                        // Default pipe timeout
 DefPipeBuffSize = 1024*64;                                                     // Define pipe buffer size

type
 TPipe = class(TObject)                                                         // Named pipe wrap class
 private
  myFile      : THandle;                                                        // Pipe file handle
  myName      : String;                                                         // Pipe name
  myServer    : Boolean;                                                        // Server or client?
  myRxFifo    : String;                                                         // Receiver    FIFO
  myTxFifo    : String;                                                         // Transmitter FIFO
  myConnected : Boolean;                                                        // Connected?
  myRxPending : Boolean;                                                        // Receiver pending IO
  myTxPending : Boolean;                                                        // Transmitter pending IO
  myRxOverlap : TOverlapped;                                                    // Receiver overlap
  myTxOverlap : TOverlapped;                                                    // Transmitter overlap
  myRxBuffer  : packed array[0..DefPipeBuffSize-1] of char;                     // Receiver buffer
  myTxBuffer  : packed array[0..DefPipeBuffSize-1] of char;                     // Transmitter buffer
  mySecDesc   : TSecurityDescriptor;                                            // Security descriptor
  mySecAttr   : TSecurityAttributes;                                            // Security attributes
  myTimeOut   : Integer;                                                        // Timeout to check connection
  myRxLost    : Int64;                                                          // Lost RxFifo bytes
  myTxLost    : Int64;                                                          // Lost TxFifo bytes
  myRxLast    : Int64;                                                          // Last value of RxLost
  myTxLast    : Int64;                                                          // Last value of TxLost
  myLastCheck : Cardinal;                                                       // Time when last check done
 private
  function    GetHost:String;
  function    GetName:String;
  function    GetHandle:THandle;
  function    GetIsServer:Boolean;
  function    GetConnected:Boolean;
  procedure   Report(const What:String; Code:Integer);
  procedure   Close(aErrorCode,aRxLost,aTxLost:Integer);
 public
  constructor Create(const aHost : String;                                      // Host(client) or empty(server)
                     const aName : String;                                      // Pipe name
                     aTimeOut    : Integer);                                    // Timeout for connections
  procedure   AfterConstruction; override;
  procedure   BeforeDestruction; override;
  destructor  Destroy; override;
 public
  procedure   Poll;                                                             // To be called by timer
  property    Host               : String    read GetHost;                      // Get host name (client)
  property    Name               : String    read GetName;                      // Get pipe name
  property    Handle             : THandle   read GetHandle;                    // Pipe handle
  property    IsServer           : Boolean   read GetIsServer;                  // Server or client
  property    Connected          : Boolean   read GetConnected;                 // Connected?
  function    GetRxFifo(Count:Integer=MaxInt):String;                           // Receiver    FIFO
  procedure   PutTxFifo(const Data:String);                                     // Transmitter FIFO
 end;

function  NewPipe(const aHost : String;                                         // Host(client) or empty(server)
                  const aName : String;                                         // Pipe name
                  aTimeOut    : Integer         = DefPipeTimeOut                // Timeout for connections
                             ): TPipe;

const FullPipeList:TList=nil;        // Full list of pipes
const FullPipeBugs:TStringList=nil;  // Full list of pipe bugs

implementation

function EasyIpc_Init(const PathName,Options:String):THandle;
var Host,Name:String;
begin
 if (Pos('\',PathName)>0) then begin
  Name:=Copy(PathName,Pos('\',PathName)+1,MaxInt);
  Host:=Copy(PathName,1,Pos('\',PathName)-1);
 end else begin
  Name:=PathName;
  Host:='';
 end;
 Result:=THandle(NewPipe(Host,Name,1000));
end;

function EasyIpc_Free(hIpc:THandle):Boolean;
begin
 Result:=false;
 if (FullPipeList=nil) then Exit;
 if FullPipeList.IndexOf(Pointer(hIpc))<0 then Exit;
 TPipe(hIpc).Free;
end;

function EasyIpc_PutText(hIpc:THandle; const Lines:String):Boolean;
begin
 Result:=false;
 if (FullPipeList=nil) then Exit;
 if FullPipeList.IndexOf(Pointer(hIpc))<0 then Exit;
 if (Lines<>'') then TPipe(hIpc).PutTxFifo(Lines);
 TPipe(hIpc).Poll;
end;

function EasyIpc_GetText(hIpc:THandle; Count:Integer):String;
begin
 Result:='';
 if (FullPipeList=nil) then Exit;
 if FullPipeList.IndexOf(Pointer(hIpc))<0 then Exit;
 TPipe(hIpc).Poll;
 if (Count>0) then Result:=TPipe(hIpc).GetRxFifo(Count);
end;

 {
 ********************
 TPipe implementation
 ********************
 }
function TPipe.GetHost:String;
begin
 Result:='';
 if not Assigned(Self) then Exit;
 if myServer then Exit;
 Result:=Copy(myName,3,Length(myName)-2);
 Delete(Result,Pos('\',Result),Length(Result));
end;

function TPipe.GetName:String;
begin
 Result:='';
 if not Assigned(Self) then Exit;
 Result:=Copy(myName,3,Length(myName)-2);
 Result:=StringReplace(Result,'\pipe\','\',[]);
end;

function TPipe.GetHandle:THandle;
begin
 if Assigned(Self) then Result:=myFile else Result:=0;
end;

function TPipe.GetIsServer:Boolean;
begin
 if Assigned(Self) then Result:=myServer else Result:=false;
end;

function TPipe.GetRxFifo(Count:Integer):String;
begin
 Result:='';
 if not Assigned(Self) then Exit;
 Count:=Min(Count,Length(myRxFifo));
 if (Count<=0) then Exit;
 Result:=Copy(myRxFifo,1,Count);
 Delete(myRxFifo,1,Count); 
end;

procedure TPipe.PutTxFifo(const Data:String);
begin
 if Assigned(Self) then begin
  myTxFifo:=myTxFifo+Data; 
 end;
end;

function TPipe.GetConnected:Boolean;
begin
 if Assigned(Self) then Result:=myConnected else Result:=false;
end;

procedure TPipe.Report(const What:String; Code:Integer);
begin
 if Assigned(Self) then if Assigned(FullPipeBugs) then begin
  FullPipeBugs.Add(FormatDateTime('yyyy.mm.dd-hh:nn:ss',Now)+' => '+myName+' '+What+' Code '+IntToStr(Code));
 end;
end;

procedure TPipe.Close(aErrorCode,aRxLost,aTxLost:Integer);
begin
 if Assigned(Self) then begin
  if myFile <> INVALID_HANDLE_VALUE then begin                                 // If file opened
   CloseHandle(myFile);                                                        // Close file
   myFile:=INVALID_HANDLE_VALUE;                                               // Mark it closed
   myConnected:=false;                                                         // No connection
   myRxPending:=false;                                                         // No pending read
   myTxPending:=false;                                                         // No pending write
   FillChar(myRxOverlap,sizeof(myRxOverlap),0);                                // Clear read overlap
   FillChar(myTxOverlap,sizeof(myTxOverlap),0);                                // Clear write overlap
  end;
  if (aErrorCode<>ERROR_SUCCESS) or (aRxLost+aTxLost<>0) then begin            // If error occured
   myRxLost:=myRxLost+Length(myRxFifo)+aRxLost;                                // Receiver data lost
   myTxLost:=myTxLost+Length(myTxFifo)+aTxLost;                                // Transmitter data lost
  end;
  Report('Close',aErrorCode);
 end;
end;

constructor TPipe.Create(const aHost : String;
                         const aName : String;
                         aTimeOut    : Integer);
begin
 inherited Create;
 myFile:=INVALID_HANDLE_VALUE;
 myServer:=(Trim(aHost)='');
 if IsServer
 then myName:=Format('\\%s\pipe\%s',['.',Trim(aName)])
 else myName:=Format('\\%s\pipe\%s',[Trim(aHost),Trim(aName)]);
 myRxFifo:='';
 myTxFifo:='';
 myConnected:=false;
 myRxPending:=false;
 myTxPending:=false;
 FillChar(myRxOverlap,sizeof(myRxOverlap),0);
 FillChar(myTxOverlap,sizeof(myTxOverlap),0);
 FillChar(myRxBuffer,sizeof(myRxBuffer),0);
 FillChar(myTxBuffer,sizeof(myTxBuffer),0);
 if InitializeSecurityDescriptor(@mySecDesc,SECURITY_DESCRIPTOR_REVISION) and
    SetSecurityDescriptorDacl(@mySecDesc,True,nil,True)
 then Report('InitializeSecurity',ERROR_SUCCESS)
 else Report('InitializeSecurity',GetLastError);
 mySecAttr.nLength:=sizeof(mySecAttr);
 mySecAttr.lpSecurityDescriptor:=@mySecDesc;
 mySecAttr.bInheritHandle:=False;
 myTimeOut:=aTimeOut;
 myRxLost:=0; myRxLast:=0;
 myTxLost:=0; myTxLast:=0;
 myLastCheck:=0;
end;

procedure TPipe.AfterConstruction;
begin
 inherited AfterConstruction;
 if (FullPipeList<>nil) then FullPipeList.Add(Self);
end;

procedure TPipe.BeforeDestruction;
begin
 if (FullPipeList<>nil) then FullPipeList.Remove(Self);
 inherited BeforeDestruction;
end;

destructor TPipe.Destroy;
begin
 Close(ERROR_SUCCESS,0,0);
 myName:='';
 myRxFifo:='';
 myTxFifo:='';
 inherited Destroy;
end;

procedure TPipe.Poll;
var RxCount,TxCount:DWORD;
 procedure PutFifoBuff(var Fifo:String; Buff:PChar; Count:Integer);
 var S:String;
 begin
  if Assigned(Buff) and (Count>0) then begin S:=''; SetString(S,Buff,Count); Fifo:=Fifo+S; S:=''; end;
 end;
begin
 if Assigned(Self) then begin
  //
  // If file handle is not created, try to create it
  //
  if myFile = INVALID_HANDLE_VALUE then begin                                  // If file not opened
   if IsServer then begin                                                      // Server:
    myFile:=CreateNamedPipe(PChar(myName),                                     // pointer to pipe name
                            PIPE_ACCESS_DUPLEX or FILE_FLAG_OVERLAPPED,        // pipe open mode
                            PIPE_TYPE_BYTE or PIPE_READMODE_BYTE,              // pipe-specific modes
                            1,                                                 // maximum number of instances
                            sizeof(myTxBuffer),                                // output buffer size, in bytes
                            sizeof(myRxBuffer),                                // input buffer size, in bytes
                            myTimeOut,                                         // time-out time, msec
                            @mySecAttr);                                       // pointer to security attributes structure
    if myFile <> INVALID_HANDLE_VALUE                                          // If file created
    then Report('CreateNamedPipe',ERROR_SUCCESS)                               // Then report success
    else Report('CreateNamedPipe',GetLastError);                               // Else report error
   end else begin                                                              // Client:
    myFile:=CreateFile(PChar(myName),                                          // pointer to name of the file
                       GENERIC_READ or GENERIC_WRITE,                          // access mode
                       FILE_SHARE_READ or FILE_SHARE_WRITE,                    // share mode
                       @mySecAttr,                                             // security
                       OPEN_EXISTING,                                          // how to create
                       FILE_FLAG_OVERLAPPED,                                   // use overlapped I/O
                       0);                                                     // template file - not used
    if myFile <> INVALID_HANDLE_VALUE                                          // If file created
    then Report('CreateFile',ERROR_SUCCESS)                                    // Report success
    else Report('CreateFile',GetLastError);                                    // Else report error
   end;
   if myFile <> INVALID_HANDLE_VALUE then begin                                // If file was created
    myConnected:=false;                                                        // Mark no connection
    myRxPending:=false;                                                        // Mark no pending read
    myTxPending:=false;                                                        // Mark no pending write
    FillChar(myRxOverlap,sizeof(myRxOverlap),0);                               // Clear read overlap
    FillChar(myTxOverlap,sizeof(myTxOverlap),0);                               // Clear write overlap
   end else Sleep(myTimeOut);                                                  // If file fails,wait some time
  end;
  //
  // File created, may connect/read/write
  //
  if myFile <> INVALID_HANDLE_VALUE then begin
   //
   // If not still connected, try to connect now...
   //
   if not myConnected then begin
    while ReadFile(myFile,myRxBuffer,sizeof(myRxBuffer),RxCount,@myRxOverlap)  // Read data to buffer
    do PutFifoBuff(myRxFifo,myRxBuffer,RxCount);                               // Transfer to FIFO
    case GetLastError of                                                       // Analize error:
     ERROR_PIPE_LISTENING:;                                                    // Server waiting for client
     ERROR_IO_PENDING:                                                         // Connection established
      begin
       myConnected:=true;                                                      // Mark we are connected
       myRxPending:=true;                                                      // Mark pending read
       Report('Connected',ERROR_SUCCESS);                                      // Mark connection
      end;
     ERROR_HANDLE_EOF: Close(GetLastError,RxCount,0);                          // End of file reached
     ERROR_BROKEN_PIPE: Close(GetLastError,RxCount,0);                         // Connection lost
     else Close(GetLastError,RxCount,0);                                       // Unknown error occured
    end;
   end;
   //
   // If connection esteblished, try to read from pipe...
   //
   if myConnected then begin                                                   // When connected
    if myRxPending then begin                                                  // When pending read:
     if GetOverlappedResult(myFile,myRxOverlap,RxCount,FALSE) then begin       // If pending read complete
      PutFifoBuff(myRxFifo,myRxBuffer,RxCount);                                // Put data to FIFO
      myRxPending:=false;                                                      // Clear pending flag
     end else                                                                  // Else
     case GetLastError of                                                      // Analize errors:
      ERROR_IO_INCOMPLETE: ;                                                   // Wait for read completion
      ERROR_HANDLE_EOF: Close(GetLastError,RxCount,0);                         // End of file reached
      ERROR_BROKEN_PIPE: Close(GetLastError,RxCount,0);                        // Connection lost
      else Close(GetLastError,RxCount,0);                                      // Unknown error occured
     end;
    end else begin                                                             // When no pending read:
     if ReadFile(myFile,myRxBuffer,RxCount,RxCount,@myRxOverlap)               // If read immediatly
     then PutFifoBuff(myRxFifo,myRxBuffer,RxCount)                             // Then put data to FIFO
     else                                                                      // If not read
     case GetLastError of                                                      // Analize error:
      ERROR_IO_PENDING: myRxPending:=true;                                     // Mark pending read
      ERROR_HANDLE_EOF: Close(GetLastError,RxCount,0);                         // End of file reached
      ERROR_BROKEN_PIPE: Close(GetLastError,RxCount,0);                        // Connection lost
      else Close(GetLastError,RxCount,0);                                      // Unknown error occured
     end;
    end;
   end;
   //
   // If connection esteblished, try to write to pipe...
   //
   if myConnected then begin                                                   // When connected
    if myTxPending then begin                                                  // When pending write:
     if GetOverlappedResult(myFile,myTxOverlap,TxCount,FALSE)                  // If write complete
     then myTxPending:=false                                                   // Clear pending flag
     else                                                                      // If not write complete
     case GetLastError of                                                      // Analize error:
      ERROR_IO_INCOMPLETE: ;                                                   // Wait for write completion
      ERROR_BROKEN_PIPE: Close(GetLastError,0,TxCount);                        // Connection lost
      else Close(GetLastError,0,TxCount);                                      // Unknown error occured
     end;
    end else begin                                                             // When no pending write:
     TxCount:=Min(Length(myTxFifo),sizeof(myTxBuffer));                        // Check data size to write
     if TxCount>0 then begin                                                   // If has data to write
      Move(PChar(myTxFifo)^,myTxBuffer,TxCount); Delete(myTxFifo,1,TxCount);   // Copy FIFO to buffer
      if not WriteFile(myFile,myTxBuffer,TxCount,TxCount,@myTxOverlap) then    // If not written immediatly
      case GetLastError of                                                     // Analize error:
       ERROR_IO_PENDING: myTxPending:=true;                                    // Mark pending write
       ERROR_BROKEN_PIPE: Close(GetLastError,0,TxCount);                       // Connection lost
       else Close(GetLastError,0,TxCount);                                     // Unknown error occured
      end;
     end;
    end;
   end;
  end;
  //
  // Check if FIFO data lost, report if so
  //
  if abs(GetTickCount-myLastCheck)>myTimeOut then begin                        // Every TimeOut milliseconds
   if myRxLost<>myRxLast then begin                                            // Check if RxLost changed
    if myRxLost>myRxLast then Report('RxLost',-1);                             // If increased, report Rx data lost
    myRxLast:=myRxLost;                                                        // Remember for future
   end;
   if myTxLost<>myTxLast then begin                                            // Check if TxLost changed
    if myTxLost>myTxLast then Report('TxLost',-1);                             // If increased, report Tx data lost
    myTxLast:=myTxLost;                                                        // Remember for future
   end;
   myLastCheck:=GetTickCount;                                                  // Remember moment of last check
  end;
 end;
end;

function  NewPipe(const aHost : String;
                  const aName : String;
                  aTimeOut    : Integer
                            ) : TPipe;
begin
 Result:=TPipe.Create(aHost,aName,aTimeOut);
end;

initialization

 FullPipeList:=TList.Create;
 FullPipeBugs:=TStringList.Create;

finalization

 FreeAndNil(FullPipeList);
 FreeAndNil(FullPipeBugs);
 
end.

